import unittest

from betago.gtp import command, response


class ResponseTestCase(unittest.TestCase):
    def setUp(self):
        self.cmd = command.Command(None, 'genmove', ('black',))
        self.cmd_with_sequence = command.Command(99, 'genmove', ('black',))

    def test_serialize(self):
        resp = response.success('D4')

        self.assertEqual('= D4\n\n', response.serialize(self.cmd, resp))

    def test_serialize_error(self):
        resp = response.error('fail!')

        self.assertEqual('? fail!\n\n', response.serialize(self.cmd, resp))

    def test_serialize_with_sequence(self):
        resp = response.success('D4')

        self.assertEqual('=99 D4\n\n', response.serialize(self.cmd_with_sequence, resp))

    def test_serialize_error_with_sequence(self):
        resp = response.error('fail!')

        self.assertEqual('?99 fail!\n\n', response.serialize(self.cmd_with_sequence, resp))



import unittest

from betago.gtp import command


class CommandTestCase(unittest.TestCase):
    def test_parse(self):
        command_string = 'play white D4'
        expected = command.Command(
            sequence=None,
            name='play',
            args=('white', 'D4'),
        )
        self.assertEqual(expected, command.parse(command_string))

    def test_parse_with_sequence_number(self):
        command_string = '999 play white D4'
        expected = command.Command(
            sequence=999,
            name='play',
            args=('white', 'D4'),
        )
        self.assertEqual(expected, command.parse(command_string))



__all__ = [
    'coords_to_gtp_position',
    'gtp_position_to_coords',
]

# 'I' is intentionally omitted.
COLS = 'ABCDEFGHJKLMNOPQRST'


def coords_to_gtp_position(coords):
    """Convert (row, col) tuple to GTP board locations.

    Example:
    >>> coords_to_gtp_position((0, 0))
    'A1'
    """
    row, col = coords
    # coords are zero-indexed, GTP is 1-indexed.
    return COLS[col] + str(row + 1)


def gtp_position_to_coords(gtp_position):
    """Convert a GTP board location to a (row, col) tuple.

    Example:
    >>> gtp_position_to_coords('A1')
    (0, 0)
    """
    col_str, row_str = gtp_position[0], gtp_position[1:]
    return (int(row_str) - 1, COLS.find(col_str))



import unittest

from betago.dataloader.goboard import GoBoard


class GoBoardTest(unittest.TestCase):
    def test_is_move_legal(self):
        board = GoBoard()
        board.apply_move('b', (4, 4))

        self.assertTrue(board.is_move_legal('b', (3, 3)))
        self.assertTrue(board.is_move_legal('w', (3, 3)))

    def test_is_move_legal_occupied(self):
        board = GoBoard()
        board.apply_move('b', (4, 4))

        self.assertFalse(board.is_move_legal('b', (4, 4)))
        self.assertFalse(board.is_move_legal('w', (4, 4)))

    def test_is_move_legal_suicide(self):
        board = GoBoard()
        # Make a ponnuki.
        board.apply_move('b', (4, 4))
        board.apply_move('b', (5, 5))
        board.apply_move('b', (6, 4))
        board.apply_move('b', (5, 3))

        # White can't move in the center.
        self.assertFalse(board.is_move_legal('w', (5, 4)))
        # But Black can.
        self.assertTrue(board.is_move_legal('b', (5, 4)))

    def test_is_move_legal_capture_not_suicide(self):
        board = GoBoard()
        # Set up a ko.
        board.apply_move('b', (4, 4))
        board.apply_move('b', (5, 5))
        board.apply_move('b', (6, 4))
        board.apply_move('b', (5, 3))
        board.apply_move('w', (4, 5))
        board.apply_move('w', (5, 6))
        board.apply_move('w', (6, 5))

        # Either can move in the center.
        self.assertTrue(board.is_move_legal('w', (5, 4)))
        self.assertTrue(board.is_move_legal('b', (5, 4)))

    def test_is_move_legal_should_not_mutate_board(self):
        board = GoBoard()
        board.apply_move('b', (4, 4))

        board.is_move_legal('b', (4, 5))

        # Check the board.
        self.assertFalse(board.is_move_on_board((4, 5)))
        # Check the go strings as well.
        group = board.go_strings[4, 4]
        self.assertEqual(1, group.get_num_stones())
        self.assertEqual(4, group.get_num_liberties())



import unittest

from betago.gtp.board import coords_to_gtp_position, gtp_position_to_coords


class GTPCoordinateTest(unittest.TestCase):
    def test_coords_to_gtp_position(self):
        self.assertEqual('A1', coords_to_gtp_position((0, 0)))
        self.assertEqual('J3', coords_to_gtp_position((2, 8)))
        self.assertEqual('B15', coords_to_gtp_position((14, 1)))

    def test_gtp_position_to_coords(self):
        self.assertEqual((0, 0), gtp_position_to_coords('A1'))
        self.assertEqual((2, 8), gtp_position_to_coords('J3'))
        self.assertEqual((14, 1), gtp_position_to_coords('B15'))



import numpy as np
from .dataloader.base_processor import GoDataProcessor, GoFileProcessor


class SevenPlaneProcessor(GoDataProcessor):
    '''
    Implementation of a Go data processor, using seven planes of 19x19 values to represent the position of
    a go board, as explained below.

    This closely reflects the representation suggested in Clark, Storkey:
    http://arxiv.org/abs/1412.3409
    '''

    def __init__(self, data_directory='data', num_planes=7, consolidate=True, use_generator=False):
        super(SevenPlaneProcessor, self).__init__(data_directory=data_directory,
                                                  num_planes=num_planes,
                                                  consolidate=consolidate,
                                                  use_generator=use_generator)

    def feature_and_label(self, color, move, go_board, num_planes):
        '''
        Parameters
        ----------
        color: color of the next person to move
        move: move they decided to make
        go_board: represents the state of the board before they moved

        Planes we write:
        0: our stones with 1 liberty
        1: our stones with 2 liberty
        2: our stones with 3 or more liberties
        3: their stones with 1 liberty
        4: their stones with 2 liberty
        5: their stones with 3 or more liberties
        6: simple ko
        '''
        row, col = move
        enemy_color = go_board.other_color(color)
        label = row * 19 + col
        move_array = np.zeros((num_planes, go_board.board_size, go_board.board_size))
        for row in range(0, go_board.board_size):
            for col in range(0, go_board.board_size):
                pos = (row, col)
                if go_board.board.get(pos) == color:
                    if go_board.go_strings[pos].liberties.size() == 1:
                        move_array[0, row, col] = 1
                    elif go_board.go_strings[pos].liberties.size() == 2:
                        move_array[1, row, col] = 1
                    elif go_board.go_strings[pos].liberties.size() >= 3:
                        move_array[2, row, col] = 1
                if go_board.board.get(pos) == enemy_color:
                    if go_board.go_strings[pos].liberties.size() == 1:
                        move_array[3, row, col] = 1
                    elif go_board.go_strings[pos].liberties.size() == 2:
                        move_array[4, row, col] = 1
                    elif go_board.go_strings[pos].liberties.size() >= 3:
                        move_array[5, row, col] = 1
                if go_board.is_simple_ko(color, pos):
                    move_array[6, row, col] = 1
        return move_array, label


class ThreePlaneProcessor(GoDataProcessor):
    '''
    Simpler version of the above processor using just three planes. This data processor uses one plane for
    stone positions of each color and one for ko.
    '''

    def __init__(self, data_directory='data', num_planes=3, consolidate=True, use_generator=False):
        super(ThreePlaneProcessor, self).__init__(data_directory=data_directory,
                                                  num_planes=num_planes,
                                                  consolidate=consolidate,
                                                  use_generator=use_generator)

    def feature_and_label(self, color, move, go_board, num_planes):
        '''
        Parameters
        ----------
        color: color of the next person to move
        move: move they decided to make
        go_board: represents the state of the board before they moved

        Planes we write:
        0: our stones
        1: their stones
        2: ko
        '''
        row, col = move
        enemy_color = go_board.other_color(color)
        label = row * 19 + col
        move_array = np.zeros((num_planes, go_board.board_size, go_board.board_size))
        for row in range(0, go_board.board_size):
            for col in range(0, go_board.board_size):
                pos = (row, col)
                if go_board.board.get(pos) == color:
                    move_array[0, row, col] = 1
                if go_board.board.get(pos) == enemy_color:
                    move_array[1, row, col] = 1
                if go_board.is_simple_ko(color, pos):
                    move_array[2, row, col] = 1
        return move_array, label


class SevenPlaneFileProcessor(GoFileProcessor):
    '''
    File processor corresponding to the above data processor. Loading all available data into memory is simply
    not feasible, and this class allows preprocessing into an efficient, binary format.
    '''
    def __init__(self, data_directory='data', num_planes=7, consolidate=True):
        super(SevenPlaneFileProcessor, self).__init__(data_directory=data_directory,
                                                      num_planes=num_planes, consolidate=consolidate)

    def store_results(self, data_file, color, move, go_board):
        '''
        Parameters
        ----------
        color: color of the next person to move
        move: move they decided to make
        go_board: represents the state of the board before they moved

        Planes we write:
        0: our stones with 1 liberty
        1: our stones with 2 liberty
        2: our stones with 3 or more liberties
        3: their stones with 1 liberty
        4: their stones with 2 liberty
        5: their stones with 3 or more liberties
        6: simple ko
        '''
        row, col = move
        enemy_color = go_board.other_color(color)
        data_file.write('GO')
        label = row * 19 + col
        data_file.write(chr(label % 256))
        data_file.write(chr(label // 256))
        data_file.write(chr(0))
        data_file.write(chr(0))
        thisbyte = 0
        thisbitpos = 0
        for plane in range(0, 7):
            for row in range(0, go_board.board_size):
                for col in range(0, go_board.board_size):
                    thisbit = 0
                    pos = (row, col)
                    if go_board.board.get(pos) == color:
                        if plane == 0 and go_board.go_strings[pos].liberties.size() == 1:
                            thisbit = 1
                        elif plane == 1 and go_board.go_strings[pos].liberties.size() == 2:
                            thisbit = 1
                        elif plane == 2 and go_board.go_strings[pos].liberties.size() >= 3:
                            thisbit = 1
                    if go_board.board.get(pos) == enemy_color:
                        if plane == 3 and go_board.go_strings[pos].liberties.size() == 1:
                            thisbit = 1
                        elif plane == 4 and go_board.go_strings[pos].liberties.size() == 2:
                            thisbit = 1
                        elif plane == 5 and go_board.go_strings[pos].liberties.size() >= 3:
                            thisbit = 1
                    if plane == 6 and go_board.is_simple_ko(color, pos):
                            thisbit = 1
                    thisbyte = thisbyte + (thisbit << (7 - thisbitpos))
                    thisbitpos = thisbitpos + 1
                    if thisbitpos == 8:
                        data_file.write(chr(thisbyte))
                        thisbitpos = 0
                        thisbyte = 0
        if thisbitpos != 0:
            data_file.write(chr(thisbyte))



from multiprocessing import Process
from flask import Flask, request, jsonify
from flask.ext.cors import CORS
import numpy as np
from .dataloader.goboard import GoBoard
from .processor import ThreePlaneProcessor


class HTTPFrontend(object):
    '''
    HTTPFrontend is a simple Flask app served on localhost:5000, exposing a REST API to predict
    go moves.
    '''
    def __init__(self, bot):
        self.bot = bot

    def start_server(self):
        ''' Start Go model server '''
        self.server = Process(target=self.start_service)
        self.server.start()

    def stop_server(self):
        ''' Terminate Go model server '''
        self.server.terminate()
        self.server.join()

    def run(self):
        ''' Run flask app'''
        app = Flask(__name__)
        CORS(app, resources={r"/prediction/*": {"origins": "*"}})
        self.app = app

        @app.route('/')
        def home():
            return 'betago'

        @app.route('/prediction', methods=['GET', 'POST'])
        def next_move():
            '''Predict next move and send to client.

            Parses the move and hands the work off to the bot.
            '''
            content = request.json
            row = content['i']
            col = content['j']
            self.bot.apply_move((row, col))

            bot_row, bot_col = self.bot.select_move()
            print('Prediction:')
            print(bot_row, bot_col)
            result = {'i': bot_row, 'j': bot_col}
            json_result = jsonify(**result)
            return json_result

        self.app.run(host='0.0.0.0', debug=True, threaded=True, use_reloader=False)


class GoModel(object):
    '''Tracks a board and selects moves.'''
    def __init__(self, model, processor):
        '''
        Parameters:
        -----------
        processor: Instance of betago.processor.GoDataLoader, e.g. SevenPlaneProcessor
        model: In principle this can be anything that can predict go moves, given data provided by the above
               processor. In practice it may very well be (an extension of) a keras model plus glue code.
        '''
        self.model = model
        self.processor = processor
        self.go_board = GoBoard(19)
        self.num_planes = processor.num_planes

    def apply_move(self, move):
        ''' Apply the human move'''
        return NotImplemented

    def select_move(self):
        ''' Select a move for the bot'''
        return NotImplemented



class KerasBot(GoModel):
    '''
    KerasBot takes top_n predictions of a keras model and tries to apply the best move. If that move is illegal,
    choose the next best, until the list is exhausted. If no more moves are left to play, continue with random
    moves until a legal move is found.
    '''

    def __init__(self, model, processor, top_n=10):
        super(KerasBot, self).__init__(model=model, processor=processor)
        self.top_n = top_n

    def apply_move(self, move):
        # Apply human move
        self.go_board.apply_move('w', move)

    def select_move(self):
        bot_color = 'b'

        # Turn the board into a feature vector.
        # The (0, 0) is for generating the label, which we ignore.
        X, label = self.processor.feature_and_label(bot_color, (0, 0), self.go_board, self.num_planes)
        X = X.reshape((1, X.shape[0], X.shape[1], X.shape[2]))

        # Generate bot move.
        found_move = False
        top_n = 10

        pred = np.squeeze(self.model.predict(X))
        top_n_pred_idx = pred.argsort()[-top_n:][::-1]
        for idx in top_n_pred_idx:
            if not found_move:
                prediction = int(idx)
                pred_row = prediction // 19
                pred_col = prediction % 19
                pred_move = (pred_row, pred_col)
                if self.go_board.is_move_legal(bot_color, pred_move):
                    found_move = True
                    self.go_board.apply_move(bot_color, pred_move)
        if not found_move:
            while not found_move:
                pred_row = np.random.randint(19)
                pred_col = np.random.randint(19)
                pred_move = (pred_row, pred_col)
                if self.go_board.is_move_legal(bot_color, pred_move):
                    found_move = True
                    self.go_board.apply_move(bot_color, pred_move)

        return pred_move


class IdiotBot(GoModel):
    '''
    Play random moves, like a good 30k bot.
    '''
    def __init__(self, model=None, processor=ThreePlaneProcessor()):
        super(IdiotBot, self).__init__(model=model, processor=processor)

    def apply_move(self, move):
        self.go_board.apply_move('w', move)

    def select_move(self):
        found_move = False
        if not found_move:
            while not found_move:
                pred_row = np.random.randint(19)
                pred_col = np.random.randint(19)
                pred_move = (pred_row, pred_col)
                if self.go_board.is_move_legal('b', pred_move):
                    found_move = True
                    self.go_board.apply_move('b', pred_move)

        return pred_move



# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.

import copy


class GoBoard(object):
    '''
    Representation of a go board. It contains "GoStrings" to represent stones and liberties. Moreover,
    the board can account for ko and handle captured stones.
    '''
    def __init__(self, board_size=19):
        '''
        Parameters
        ----------
        ko_last_move_num_captured: How many stones have been captured last move. If this is not 1, it can't be ko.
        ko_last_move: board position of the ko.
        board_size: Side length of the board, defaulting to 19.
        go_strings: Dictionary of go_string objects representing stones and liberties.
        '''
        self.ko_last_move_num_captured = 0
        self.ko_last_move = -3
        self.board_size = board_size
        self.board = {}
        self.go_strings = {}

    def fold_go_strings(self, target, source, join_position):
        ''' Merge two go strings by joining their common moves'''
        if target == source:
            return
        for stone_position in source.stones.stones:
            self.go_strings[stone_position] = target
            target.insert_stone(stone_position)
        target.copy_liberties_from(source)
        target.remove_liberty(join_position)

    def add_adjacent_liberty(self, pos, go_string):
        '''
        Append new liberties to provided GoString for the current move
        '''
        row, col = pos
        if row < 0 or col < 0 or row > self.board_size - 1 or col > self.board_size - 1:
            return
        if pos not in self.board:
            go_string.insert_liberty(pos)

    def is_move_on_board(self, move):
        return move in self.board

    def is_move_suicide(self, color, pos):
        '''Check if a proposed move would be suicide.'''
        # Make a copy of ourself to apply the move.
        temp_board = copy.deepcopy(self)
        temp_board.apply_move(color, pos)
        new_string = temp_board.go_strings[pos]
        return new_string.get_num_liberties() == 0

    def is_move_legal(self, color, pos):
        '''Check if a proposed moved is legal.'''
        return (not self.is_move_on_board(pos)) and not (self.is_move_suicide(color, pos))

    def create_go_string(self, color, pos):
        ''' Create GoString from current Board and move '''
        go_string = GoString(self.board_size, color)
        go_string.insert_stone(pos)
        self.go_strings[pos] = go_string
        self.board[pos] = color

        row, col = pos
        for adjpos in [(row-1, col), (row+1, col), (row, col-1), (row, col+1)]:
            self.add_adjacent_liberty(adjpos, go_string)
        return go_string

    def other_color(self, color):
        '''
        Color of other player
        '''
        if color == 'b':
            return 'w'
        if color == 'w':
            return 'b'

    def is_simple_ko(self, play_color, pos):
        '''
        Determine ko from board position and player.

        Parameters:
        -----------
        play_color: Color of the player to make the next move.
        pos: Current move as (row, col)
        '''
        enemy_color = self.other_color(play_color)
        row, col = pos
        if self.ko_last_move_num_captured == 1:
            last_move_row, last_move_col = self.ko_last_move
            manhattan_distance_last_move = abs(last_move_row - row) + abs(last_move_col - col)
            if manhattan_distance_last_move == 1:
                last_go_string = self.go_strings.get((last_move_row, last_move_col))
                if last_go_string is not None and last_go_string.get_num_liberties() == 1:
                    if last_go_string.get_num_stones() == 1:
                        num_adjacent_enemy_liberties = 0
                        for adjpos in [(row-1, col), (row+1, col), (row, col-1), (row, col+1)]:
                            if (self.board.get(adjpos) == enemy_color and
                               self.go_strings[adjpos].get_num_liberties() == 1):
                                num_adjacent_enemy_liberties = num_adjacent_enemy_liberties + 1
                        if num_adjacent_enemy_liberties == 1:
                            return True
        return False

    def check_enemy_liberty(self, play_color, enemy_pos, our_pos):
        '''
        Update surrounding liberties on board after a move has been played.

        Parameters:
        -----------
        play_color: Color of player about to move
        enemy_pos: latest enemy move
        our_pos: our latest move
        '''
        enemy_row, enemy_col = enemy_pos
        our_row, our_col = our_pos

        # Sanity checks
        if enemy_row < 0 or enemy_row >= self.board_size or enemy_col < 0 or enemy_col >= self.board_size:
            return
        enemy_color = self.other_color(play_color)
        if self.board.get(enemy_pos) != enemy_color:
            return
        enemy_string = self.go_strings[enemy_pos]
        if enemy_string is None:
            raise('check_string is None')

        # Update adjacent liberties on board
        enemy_string.remove_liberty(our_pos)
        if enemy_string.get_num_liberties() == 0:
            for enemy_pos in enemy_string.stones.stones:
                string_row, string_col = enemy_pos
                del self.board[enemy_pos]
                del self.go_strings[enemy_pos]
                self.ko_last_move_num_captured = self.ko_last_move_num_captured + 1
                for adjstring in [(string_row-1, string_col), (string_row+1, string_col),
                                  (string_row, string_col-1), (string_row, string_col+1)]:
                    self.add_liberty_to_adjacent_string(adjstring, enemy_pos, play_color)

    def apply_move(self, play_color, pos):
        '''
        Execute move for given color, i.e. play current stone on this board
        Parameters:
        -----------
        play_color: Color of player about to move
        pos: Current move as (row, col)
        '''
        if pos in self.board:
            raise('>>> Error: move ' + str(pos) + 'is already on board.')

        self.ko_last_move_num_captured = 0
        row, col = pos

        # Remove any enemy stones that no longer have a liberty
        self.check_enemy_liberty(play_color, (row - 1, col), pos)
        self.check_enemy_liberty(play_color, (row + 1, col), pos)
        self.check_enemy_liberty(play_color, (row, col - 1), pos)
        self.check_enemy_liberty(play_color, (row, col + 1), pos)

        # Create a GoString for our new stone, and merge with any adjacent strings
        play_string = self.create_go_string(play_color, pos)
        play_string = self.fold_our_moves(play_string, play_color, (row - 1, col), pos)
        play_string = self.fold_our_moves(play_string, play_color, (row + 1, col), pos)
        play_string = self.fold_our_moves(play_string, play_color, (row, col - 1), pos)
        play_string = self.fold_our_moves(play_string, play_color, (row, col + 1), pos)

        # Store last move for ko
        self.ko_last_move = pos

    def add_liberty_to_adjacent_string(self, string_pos, liberty_pos, color):
        ''' Insert liberty into corresponding GoString '''
        if self.board.get(string_pos) != color:
            return
        go_string = self.go_strings[string_pos]
        go_string.insert_liberty(liberty_pos)

    def fold_our_moves(self, first_string, color, pos, join_position):
        ''' Fold current board situation with a new move played by us'''
        row, col = pos
        if row < 0 or row >= self.board_size or col < 0 or col >= self.board_size:
            return first_string
        if self.board.get(pos) != color:
            return first_string
        string_to_fold = self.go_strings[pos]
        self.fold_go_strings(string_to_fold, first_string, join_position)
        return string_to_fold

    def __str__(self):
        result = 'GoBoard\n'
        for i in range(self.board_size - 1, -1, -1):
            line = ''
            for j in range(0, self.board_size):
                thispiece = self.board.get((i, j))
                if thispiece is None:
                    line = line + '.'
                if thispiece == 'b':
                    line = line + '*'
                if thispiece == 'w':
                    line = line + 'O'
            result = result + line + '\n'
        return result


class BoardSequence(object):
    '''
    Store a sequence of locations on a board, which could either represent stones or liberties.
    '''
    def __init__(self, board_size=19):
        self.board_size = board_size
        self.stones = []
        self.board = {}

    def insert(self, combo):
        row, col = combo
        if combo in self.board:
            return
        self.stones.append(combo)
        self.board[combo] = len(self.stones) - 1

    def erase(self, combo):
        if combo not in self.board:
            return
        iid = self.board[combo]
        if iid == len(self.stones) - 1:
            del self.stones[iid]
            del self.board[combo]
            return
        self.stones[iid] = self.stones[len(self.stones) - 1]
        del self.stones[len(self.stones) - 1]
        movedcombo = self.stones[iid]
        self.board[movedcombo] = iid
        del self.board[combo]

    def exists(self, combo):
        return combo in self.board

    def size(self):
        return len(self.stones)

    def __getitem__(self, iid):
        return self.stones[iid]

    def __str__(self):
        result = 'BoardSequence\n'
        for row in range(self.board_size - 1, -1, -1):
            thisline = ""
            for col in range(0, self.board_size):
                if self.exists((row, col)):
                    thisline = thisline + "*"
                else:
                    thisline = thisline + "."
            result = result + thisline + "\n"
        return result


class GoString(object):
    '''
    Represents a string of contiguous stones of one color on the board, including a list of all its liberties.
    '''
    def __init__(self, board_size, color):
        self.board_size = board_size
        self.color = color
        self.liberties = BoardSequence(board_size)
        self.stones = BoardSequence(board_size)

    def get_stone(self, index):
        return self.stones[index]

    def get_liberty(self, index):
        return self.liberties[index]

    def insert_stone(self, combo):
        self.stones.insert(combo)

    def get_num_stones(self):
        return self.stones.size()

    def remove_liberty(self, combo):
        self.liberties.erase(combo)

    def get_num_liberties(self):
        return self.liberties.size()

    def insert_liberty(self, combo):
        self.liberties.insert(combo)

    def copy_liberties_from(self, source):
        for libertyPos in source.liberties.stones:
            self.liberties.insert(libertyPos)

    def __str__(self):
        result = "go_string[ stones=" + str(self.stones) + " liberties=" + str(self.liberties) + " ]"
        return result



# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import print_function
import os
import random
from .index_processor import KGSIndex


class Sampler(object):
    '''
    Sample training and test data from zipped sgf files such that test data is kept stable.
    '''

    def __init__(self, data_dir='data', num_test_games=100, cap_year=2015, seed=1337):
        self.data_dir = data_dir
        self.num_test_games = num_test_games
        self.test_games = []
        self.train_games = []
        self.test_folder = 'test_samples.py'
        self.cap_year = 2015

        random.seed(seed)
        self.compute_test_samples()

    def draw_samples(self, num_sample_games):
        '''
        Draw num_sample_games many training games from index.
        '''
        available_games = []
        index = KGSIndex(data_directory=self.data_dir)

        for fileinfo in index.file_info:
            filename = fileinfo['filename']
            year = int(filename.split('-')[1].split('_')[0])
            if year > self.cap_year:
                continue
            num_games = fileinfo['num_games']
            for i in range(num_games):
                available_games.append((filename, i))
        print('>>> Total number of games used: ' + str(len(available_games)))

        sample_set = set()
        while len(sample_set) < num_sample_games:
            sample = random.choice(available_games)
            if sample not in sample_set:
                sample_set.add(sample)
        print('Drawn ' + str(num_sample_games) + ' samples:')
        return list(sample_set)

    def draw_training_games(self):
        '''
        Get list of all non-test games, that are no later than dec 2014
        Ignore games after cap_year to keep training data stable
        '''
        index = KGSIndex(data_directory=self.data_dir)
        for file_info in index.file_info:
            filename = file_info['filename']
            year = int(filename.split('-')[1].split('_')[0])
            if year > self.cap_year:
                continue
            numgames = file_info['num_games']
            for i in range(numgames):
                sample = (filename, i)
                if sample not in self.test_games:
                    self.train_games.append(sample)
        print('total num training games: ' + str(len(self.train_games)))

    def compute_test_samples(self):
        '''
        If not already existing, create local file to store fixed set of test samples
        '''
        if not os.path.isfile(self.test_folder):
            test_games = self.draw_samples(self.num_test_games)
            test_sample_file = open(self.test_folder, 'w')
            for sample in test_games:
                test_sample_file.write(str(sample) + "\n")
            test_sample_file.close()

        test_sample_file = open(self.test_folder, 'r')
        sample_contents = test_sample_file.read()
        test_sample_file.close()
        for line in sample_contents.split('\n'):
            if line != "":
                (filename, index) = eval(line)
                self.test_games.append((filename, index))

    def draw_training_samples(self, num_sample_games):
        '''
        Draw training games, not overlapping with any of the test games.
        '''
        available_games = []
        index = KGSIndex(data_directory=self.data_dir)
        for fileinfo in index.file_info:
            filename = fileinfo['filename']
            year = int(filename.split('-')[1].split('_')[0])
            if year > self.cap_year:
                continue
            numgames = fileinfo['num_games']
            for i in range(numgames):
                available_games.append((filename, i))
        print('total num games: ' + str(len(available_games)))

        sample_set = set()
        while len(sample_set) < num_sample_games:
            sample = random.choice(available_games)
            if sample not in self.test_games:
                sample_set.add(sample)
        print('Drawn ' + str(num_sample_games) + ' samples:')
        return list(sample_set)

    def draw_all_training(self):
        '''
        Draw all available training games.
        '''
        available_games = []
        index = KGSIndex(data_directory=self.data_dir)

        for fileinfo in index.file_info:
            filename = fileinfo['filename']
            year = int(filename.split('-')[1].split('_')[0])
            if year > self.cap_year:
                continue
            numgames = fileinfo['numGames']
            for i in range(numgames):
                available_games.append((filename, i))
        print('total num games: ' + str(len(available_games)))

        sample_set = set()
        for sample in available_games:
            if sample not in self.test_samples:
                sample_set.add(sample)
        print('Drawn all samples, ie ' + str(len(sample_set)) + ' samples:')
        return list(sample_set)



# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import print_function
import os
import gc
import glob
import os.path
import zipfile
import numpy as np
import argparse
import multiprocessing
from os import sys
from gomill import sgf as gosgf
from keras.utils import np_utils

from .index_processor import KGSIndex
from .sampling import Sampler
from .goboard import GoBoard


def worker(jobinfo):
    try:
        clazz, dir_name, num_planes, zip_file, data_file_name, game_list = jobinfo
        clazz(dir_name, num_planes).process_zip(dir_name, zip_file, data_file_name, game_list)
    except (KeyboardInterrupt, SystemExit):
        raise('>>> Exiting child process.')


class DataGenerator(object):
    def __init__(self, data_dir, samples):
        self.data_dir = data_dir
        self.files = set(file_name for file_name, index in samples)
        self.samples = samples
        self.num_samples = None

    def get_num_samples(self, batch_size=128, nb_classes=19*19):
        if self.num_samples is not None:
            return self.num_samples
        else:
            self.num_samples = 0
            for X, y in self._generate(batch_size=batch_size, nb_classes=nb_classes):
                self.num_samples += X.shape[0]
            return self.num_samples

    def _generate(self, batch_size, nb_classes):
        for zip_file_name in self.files:
            file_name = zip_file_name.replace('.zip', '') + 'train'
            base = self.data_dir + '/' + file_name + '_features_*.npy'
            for feature_file in glob.glob(base):
                label_file = feature_file.replace('features', 'labels')
                X = np.load(feature_file)
                y = np.load(label_file)
                X = X.astype('float32')
                y = np_utils.to_categorical(y.astype(int), nb_classes)
                gc.collect()
                while X.shape[0] >= batch_size:
                    X_batch, X = X[:batch_size], X[batch_size:]
                    y_batch, y = y[:batch_size], y[batch_size:]
                    gc.collect()
                    yield X_batch, y_batch
            gc.collect()

    def generate(self, batch_size=128, nb_classes=19*19):
        while True:
            for item in self._generate(batch_size=batch_size, nb_classes=nb_classes):
                yield item


class GoBaseProcessor(object):
    '''
    Abstract base class for Go data processing. To implement this class, implement
    process_zip and consolidate_games below.

    Generally speaking, processing load is split between available CPUs, thereby generating intermediate
    files for each worker. Processing many files can produce massive amounts of data that will not (easily)
    fit into memory. To avoid overflow, deactivate file consolidation by initializing with consolidate=False.
    '''

    def __init__(self, data_directory='data', num_planes=7, consolidate=True):
        '''
        Parameters:
        -----------
        data_dir: relative path to store data files, defaults to 'data'
        num_planes: Number of Go board copies used as input, corresponds to the number of input channels
                    in a CNN.
        consolidate: Boolean flag to indicate if intermediate results should be consolidated into one, which
                     can be very expensive.
        '''
        self.data_dir = data_directory
        self.num_planes = num_planes
        self.consolidate = consolidate

    def process_zip(self, dir_name, zip_file_name, data_file_name, game_list):
        '''
        Method to determine how to process each zip file with games.
        '''
        return NotImplemented

    def consolidate_games(self, name, samples):
        '''
        Consolidate a list of results into one.
        '''
        return NotImplemented

    def load_go_data(self, types=['train'], data_dir='data', num_samples=1000):
        '''
        Main method to load go data.

        Loads and initializes an index from KGS and downloads zipped files. Depending on provided type, unzip and
        process data, then optionally store it in one consolidated file.

        Parameters:
        -----------
        types: Provide a list, subset of ['train', 'test'].
        data_dir: local folder to store data, provided as relative path.
        num_samples: Number of Go games to load.
        '''
        index = KGSIndex(data_directory=self.data_dir)
        index.download_files()

        for name in types:
            sampler = Sampler(data_dir=self.data_dir)
            if name == 'test':
                samples = sampler.test_games
            elif name == 'train' and num_samples is not None:
                samples = sampler.draw_training_samples(num_samples)
            elif name == 'train' and num_samples is None:
                samples = sampler.draw_all_training()

            # Map load to CPUs, then consolidate all examples
            self.map_to_workers(name, samples)

            # If consolidate flag is True, consolidate. Note that merging all data into, e.g.
            # one numpy array will be too expensive at some point.
            if self.consolidate:
                features_and_labels = self.consolidate_games(name, samples)
            else:
                print('>>> No consolidation done, single files stored in data folder')
        print('>>> Finished processing')
        if self.consolidate:
            return features_and_labels
        else:
            return

    def load_go_data_cli(self):
        '''
        Main method for loading Go game data from command line.
        Data types to choose from are:
            test: Load test data. This set is fixed and not contained in any train data.
            train: Load train data for specified number of games
        '''
        # Parse command line arguments
        parser = argparse.ArgumentParser(description='Process Go Game data')
        parser.add_argument('--type', dest='type', type=str, nargs='+',
                            help='Choose data type from test and train')
        parser.add_argument('--data', dest='data_dir', type=str, nargs='?',
                            help='Optional data directory')
        parser.add_argument('--n', dest='num_samples', type=int, nargs='?',
                            help='Number of samples used in training. If none provided, load full data set')
        args = parser.parse_args()

        # Determine data types for training
        types = args.type
        for t in types:
            if t not in ['test', 'train']:
                raise("""Unsupported type, choose from 'test', 'train' """)

        # Determine number of training samples
        num_samples = None if args.num_samples is None else args.num_samples

        # Determine local directory to put all data
        target_dir = 'data' if args.data_dir is None else args.data_dir
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)
        self.data_dir = target_dir

        self.load_go_data(types=types, data_dir=self.data_dir, num_samples=num_samples)

    def get_handicap(self, go_board, sgf):
        ''' Get handicap stones '''
        first_move_done = False
        if sgf.get_handicap() != None and sgf.get_handicap() != 0:
            for setup in sgf.get_root().get_setup_stones():
                for move in setup:
                    go_board.apply_move('b', move)
            first_move_done = True
        return go_board, first_move_done

    def map_to_workers(self, name, samples):
        '''
        Determine the list of zip files that need to be processed, then map load
        to number of available CPUs.
        '''
        # Create a dictionary with file names as keys and games as values
        zip_names = set()
        indices_by_zip_name = {}
        for filename, index in samples:
            zip_names.add(filename)
            if filename not in indices_by_zip_name:
                indices_by_zip_name[filename] = []
            indices_by_zip_name[filename].append(index)
        print('>>> Number of zip files: ' + str(len(zip_names)))

        # Transform the above dictionary to a list that can be processed in parallel
        zips_to_process = []
        for zip_name in zip_names:
            base_name = zip_name.replace('.zip', '')
            data_file_name = base_name + name
            if not os.path.isfile(self.data_dir + '/' + data_file_name):
                zips_to_process.append((self.__class__, self.data_dir, self.num_planes, zip_name,
                                        data_file_name, indices_by_zip_name[zip_name]))

        # Determine number of CPU cores and split work load among them
        cores = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=cores)
        p = pool.map_async(worker, zips_to_process)
        try:
            results = p.get(0xFFFF)
            print(results)
        except KeyboardInterrupt:
            print("Caught KeyboardInterrupt, terminating workers")
            pool.terminate()
            pool.join()
            sys.exit(-1)

    def init_go_board(self, sgf_contents):
        ''' Initialize a 19x19 go board from SGF file content'''
        sgf = gosgf.Sgf_game.from_string(sgf_contents)
        return sgf, GoBoard(19)

    def num_total_examples(self, this_zip, game_list, name_list):
        ''' Total number of moves, i.e. training samples'''
        total_examples = 0
        for index in game_list:
            name = name_list[index + 1]
            if name.endswith('.sgf'):
                content = this_zip.read(name)
                sgf, go_board_no_handy = self.init_go_board(content)
                go_board, first_move_done = self.get_handicap(go_board_no_handy, sgf)

                num_moves = 0
                for item in sgf.main_sequence_iter():
                    color, move = item.get_move()
                    if color is not None and move is not None:
                        if first_move_done:
                            num_moves = num_moves + 1
                        first_move_done = True
                total_examples = total_examples + num_moves
            else:
                raise(name + ' is not a valid sgf')
        return total_examples


class GoDataProcessor(GoBaseProcessor):
    '''
    GoDataProcessor generates data, e.g. numpy arrays, of features and labels and returns them to the user.
    '''
    def __init__(self, data_directory='data', num_planes=7, consolidate=True, use_generator=False):
        super(GoDataProcessor, self).__init__(data_directory=data_directory,
                                              num_planes=num_planes, consolidate=consolidate)
        self.use_generator = use_generator

    def feature_and_label(self, color, move, go_board):
        '''
        Given a go board and the next move for a given color, treat the next move as label and
        process the current board situation as feature to learn this label.

        return: X, y - feature and label
        '''
        return NotImplemented

    def process_zip(self, dir_name, zip_file_name, data_file_name, game_list):
        # Read zipped file and extract name list
        this_zip = zipfile.ZipFile(dir_name + '/' + zip_file_name)
        name_list = this_zip.namelist()

        # Determine number of examples
        total_examples = self.num_total_examples(this_zip, game_list, name_list)

        features = np.zeros((total_examples, self.num_planes, 19, 19))
        labels = np.zeros((total_examples,))

        counter = 0
        for index in game_list:
            name = name_list[index + 1]
            if name.endswith('.sgf'):
                '''
                Load Go board and determine handicap of game, then iterate through all moves,
                store preprocessed move in data_file and apply move to board.
                '''
                sgf_content = this_zip.read(name)
                sgf, go_board_no_handy = self.init_go_board(sgf_content)
                go_board, first_move_done = self.get_handicap(go_board_no_handy, sgf)
                for item in sgf.main_sequence_iter():
                    color, move = item.get_move()
                    if color is not None and move is not None:
                        row, col = move
                        if first_move_done:
                            X, y = self.feature_and_label(color, move, go_board, self.num_planes)
                            features[counter] = X
                            labels[counter] = y
                            counter += 1
                        go_board.apply_move(color, (row, col))
                        first_move_done = True
            else:
                raise(name + ' is not a valid sgf')

        feature_file_base = dir_name + '/' + data_file_name + '_features_%d'
        label_file_base = dir_name + '/' + data_file_name + '_labels_%d'

        # Due to files with large content, split up after chunksize
        chunk = 0
        chunksize = 1024
        while features.shape[0] >= chunksize:
            feature_file = feature_file_base % chunk
            label_file = label_file_base % chunk
            chunk += 1
            cur_features, features = features[:chunksize], features[chunksize:]
            cur_labels, labels = labels[:chunksize], labels[chunksize:]
            np.save(feature_file, cur_features)
            np.save(label_file, cur_labels)

    def consolidate_games(self, name, samples):
        print('>>> Creating consolidated numpy arrays')

        if self.use_generator:
            print('>>> Return generator')
            generator = DataGenerator(self.data_dir, samples)
            return generator

        files_needed = set(file_name for file_name, index in samples)
        print('>>> Total number of files: ' + str(len(files_needed)))

        file_names = []
        for zip_file_name in files_needed:
            file_name = zip_file_name.replace('.zip', '') + name
            file_names.append(file_name)

        feature_list = []
        label_list = []
        for file_name in file_names:
            X = np.load(self.data_dir + '/' + file_name + '_features.npy')
            y = np.load(self.data_dir + '/' + file_name + '_labels.npy')
            feature_list.append(X)
            label_list.append(y)
            print('>>> Done')

        features = np.concatenate(feature_list, axis=0)
        labels = np.concatenate(label_list, axis=0)

        feature_file = self.data_dir + '/' + str(self.num_planes) + '_plane_features_' + name
        label_file = self.data_dir + '/' + str(self.num_planes) + '_plane_labels_' + name

        np.save(feature_file, features)
        np.save(label_file, labels)

        return features, labels


class GoFileProcessor(GoBaseProcessor):
    '''
    GoFileProcessor generates files, e.g. binary representations, of features and labels.
    '''
    def __init__(self, data_directory='data', num_planes=7, consolidate=True):
        super(GoFileProcessor, self).__init__(data_directory=data_directory,
                                              num_planes=num_planes, consolidate=consolidate)

    def store_results(self, data_file, color, move, go_board):
        '''
        Apply current move of given color to provided board and store to data_file.
        '''
        return NotImplemented

    def write_file_header(self, data_file, n, num_planes, board_size, bits_per_pixel):
        headerLine = 'mlv2'
        headerLine = headerLine + '-n=' + str(n)
        headerLine = headerLine + '-num_planes=' + str(num_planes)
        headerLine = headerLine + '-imagewidth=' + str(board_size)
        headerLine = headerLine + '-imageheight=' + str(board_size)
        headerLine = headerLine + '-datatype=int'
        headerLine = headerLine + '-bpp=' + str(bits_per_pixel)
        print(headerLine)
        headerLine = headerLine + "\0\n"
        headerLine = headerLine + chr(0) * (1024-len(headerLine))
        data_file.write(headerLine)

    def process_zip(self, dir_name, zip_file_name, data_file_name, game_list):
        # Read zipped file and extract name list
        this_zip = zipfile.ZipFile(dir_name + '/' + zip_file_name)
        name_list = this_zip.namelist()

        # Determine number of examples
        total_examples = self.num_total_examples(this_zip, game_list, name_list)
        print('>>> Total number of Go games in this zip: ' + str(total_examples))

        # Write file header
        data_file = open(dir_name + '/' + data_file_name, 'wb')
        self.write_file_header(data_file=data_file, n=total_examples, num_planes=7, board_size=19, bits_per_pixel=1)

        # Write body and close file
        for index in game_list:
            name = name_list[index + 1]
            if name.endswith('.sgf'):
                '''
                Load Go board and determine handicap of game, then iterate through all moves,
                store preprocessed move in data_file and apply move to board.
                '''
                sgf_content = this_zip.read(name)
                sgf, go_board_no_handy = self.init_go_board(sgf_content)
                go_board, first_move_done = self.get_handicap(go_board_no_handy, sgf)
                move_idx = 0
                for item in sgf.main_sequence_iter():
                    (color, move) = item.get_move()
                    if color is not None and move is not None:
                        row, col = move
                        if first_move_done:
                            self.store_results(data_file, color, move, go_board)
                        go_board.apply_move(color, (row, col))
                        move_idx = move_idx + 1
                        first_move_done = True
            else:
                raise(name + ' is not a valid sgf')
        data_file.write('END')
        data_file.close()

    def consolidate_games(self, name, samples):
        print('>>> Creating consolidated .dat...')
        file_path = self.data_dir + '/kgsgo_' + name
        if os.path.isfile(file_path):
            print('>>> File ' + file_path + ' already exists')
            return

        files_needed = set(file_name for file_name, index in samples)
        print('>>> Total dat files to be consolidated: ' + str(len(files_needed)))

        # Collect names of data files
        data_file_names = []
        for zip_file_name in files_needed:
            data_file_name = zip_file_name.replace('.zip', '') + name
            data_file_names.append(data_file_name)

        # Count total number of moves
        num_records = 0
        for data_file_name in data_file_names:
            if not os.path.isfile(self.data_dir + '/' + data_file_name):
                print('>>> Missing file: ' + data_file_name)
                sys.exit(-1)
            child = open(self.data_dir + '/' + data_file_name, 'rb')
            header = child.read(1024)
            this_n = int(header.split('-n=')[1].split('-')[0])
            child.close()
            num_records = num_records + this_n

        # Write content to consolidate file
        consolidated_file = open(file_path, 'wb')
        self.write_file_header(consolidated_file, num_records, self.num_planes, 19, 1)
        for filename in data_file_names:
            print('>>> Reading from ' + filename + ' ...')
            file_path = self.data_dir + '/' + filename
            single_dat = open(file_path, 'rb')
            single_dat.read(1024)
            data = single_dat.read()
            if data[-3:] != 'END':
                raise('Invalid file, doesnt end with END: ' + file_path)
            consolidated_file.write(data[:-3])
            single_dat.close()
        consolidated_file.write('END')
        consolidated_file.close()



__all__ = [
    'Command',
]


class Command(object):
    """A GTP command.

    A GTP command contains:
    - An optional sequence number (used for matching up responses with
      commands)
    - A command name, e.g. 'genmove'
    - One or more arguments to the command, e.g. 'black'
    """
    def __init__(self, sequence, name, args):
        self.sequence = sequence
        self.name = name
        self.args = tuple(args)

    def __eq__(self, other):
        return self.sequence == other.sequence and \
            self.name == other.name and \
            self.args == other.args

    def __repr__(self):
        return 'Command(%r, %r, %r)' % (self.sequence, self.name, self.args)

    def __str__(self):
        return repr(self)


def parse(command_string):
    """Parse a GTP protocol line into a Command object.

    Example:
    >>> parse('999 play white D4')
    Command(999, 'play', ('white', 'D4'))
    """
    pieces = command_string.split()
    # Check for the sequence number.
    try:
        sequence = int(pieces[0])
        pieces = pieces[1:]
    except ValueError:
        # The first piece was non-numeric, so there was no sequence
        # number.
        sequence = None
    name, args = pieces[0], pieces[1:]
    return Command(sequence, name, args)



import sys

from . import command, response
from .board import *

__all__ = [
    'GTPFrontend',
]


class GTPFrontend(object):
    """Go Text Protocol frontend for a bot.

    Handles parsing GTP commands and formatting responses.

    Extremely limited implementation right now:
     - Only supports 19x19 boards.
     - Doesn't support handicap.
     - Can only play as black.
     - When white passes, black will pass too.
    """
    def __init__(self, bot):
        self.bot = bot
        self._input = sys.stdin
        self._output = sys.stdout
        self._stopped = False
        self._will_pass = False

    def run(self):
        while not self._stopped:
            ln = self._input.readline().strip()
            cmd = command.parse(ln)
            resp = self.process(cmd)
            self._output.write(response.serialize(cmd, resp))
            self._output.flush()

    def process(self, command):
        handlers = {
            'boardsize': self.handle_boardsize,
            'clear_board': self.ignore,
            'genmove': self.handle_genmove,
            'known_command': self.handle_known_command,
            'komi': self.ignore,
            'play': self.handle_play,
            'quit': self.handle_quit,
        }
        handler = handlers.get(command.name, self.handle_unknown)
        resp = handler(*command.args)
        return resp

    def ignore(self, *args):
        """Placeholder for commands we haven't dealt with yet."""
        return response.success()

    def handle_known_command(self, command_name):
        # TODO Should actually check if the command is known.
        return response.success('false')

    def handle_play(self, player, move):
        if player != 'white':
            return Error('Can only play as black')
        if move == 'pass':
            self._will_pass = True
        else:
            self.bot.apply_move(gtp_position_to_coords(move))
        return response.success()

    def handle_genmove(self, player):
        if player != 'black':
            return response.error('Can only play as black')
        # TODO The bot should decide when the pass.
        if self._will_pass:
            return response.success('pass')
        move = self.bot.select_move()
        return response.success(coords_to_gtp_position(move))

    def handle_boardsize(self, size):
        if int(size) != 19:
            return response.error('Only 19x19 currently supported')
        return response.success()

    def handle_quit(self):
        self._stopped = True
        return response.success()

    def handle_unknown(self, *args):
        return response.error('Unrecognized command')



# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import print_function, unicode_literals
from builtins import unicode
import os
import sys
import urllib
import multiprocessing


def worker(url_and_target):
    '''
    Parallelize data download via multiprocessing
    '''
    try:
        (url, target_path) = url_and_target
        print('>>> Downloading ' + target_path)
        urllib.urlretrieve(url, target_path)
    except (KeyboardInterrupt, SystemExit):
        print('>>> Exiting child process')


class KGSIndex(object):

    def __init__(self,
                 kgs_url='http://u-go.net/gamerecords/',
                 index_page='kgs_index.html',
                 data_directory='data'):
        '''
        Create an index of zip files containing SGF data of actual Go Games on KGS.

        Parameters:
        -----------
        kgs_url: URL with links to zip files of games
        index_page: Name of local html file of kgs_url
        data_directory: name of directory relative to current path to store SGF data
        '''
        self.kgs_url = kgs_url
        self.index_page = index_page
        self.data_directory = data_directory
        self.file_info = []
        self.urls = []
        self.load_index()  # Load index on creation

    def download_files(self):
        '''
        Download zip files by distributing work on all available CPUs
        '''
        if not os.path.isdir(self.data_directory):
            os.makedirs(self.data_directory)

        urls_to_download = []
        for file_info in self.file_info:
            url = file_info['url']
            file_name = file_info['filename']
            if not os.path.isfile(self.data_directory + '/' + file_name):
                urls_to_download.append((url, self.data_directory + '/' + file_name))
        cores = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=cores)
        try:
            it = pool.imap(worker, urls_to_download)
            for i in it:
                pass
            pool.close()
            pool.join()
        except KeyboardInterrupt:
            print(">>> Caught KeyboardInterrupt, terminating workers")
            pool.terminate()
            pool.join()
            sys.exit(-1)

    def create_index_page(self):
        '''
        If there is no local html containing links to files, create one.
        '''
        if os.path.isfile(self.index_page):
            print('>>> Reading cached index page')
            index_file = open(self.index_page, 'r')
            index_contents = index_file.read()
            index_file.close()
        else:
            print('>>> Downloading index page')
            fp = urllib.urlopen(self.kgs_url)
            data = unicode(fp.read())
            fp.close()
            index_contents = data
            index_file = open(self.index_page, 'w')
            index_file.write(index_contents)
            index_file.close()
        return index_contents

    def load_index(self):
        '''
        Create the actual index representation from the previously downloaded or cached html.
        '''
        index_contents = self.create_index_page()
        split_page = [item for item in index_contents.split('<a href="') if item.startswith("http://")]
        for item in split_page:
            download_url = item.split('">Download')[0]
            if download_url.endswith('.zip'):
                self.urls.append(download_url)
        for url in self.urls:
            filename = os.path.basename(url)
            split_file_name = filename.split('-')
            num_games = int(split_file_name[len(split_file_name)-2])
            print(filename + ' ' + str(num_games))
            self.file_info.append({'url': url, 'filename': filename, 'num_games': num_games})


if __name__ == '__main__':
    index = KGSIndex()
    index.download_files()



from setuptools import setup
from setuptools import find_packages

setup(name='betago',
      version='0.2',
      description='Create your own go bot from scratch. We are Lee Sedol!',
      url='http://github.com/maxpumperla/betago',
      download_url='https://github.com/maxpumperla/betago/tarball/0.2',
      author='Max Pumperla',
      author_email='max.pumperla@googlemail.com',
      install_requires=['keras', 'gomill', 'flask', 'flask-cors', 'future', 'h5py'],
      license='MIT',
      packages=find_packages(),
      zip_safe=False)



__all__ = [
    'Response',
    'error',
    'serialize',
    'success',
]


class Response(object):
    """Response to a GTP command."""
    def __init__(self, status, body):
        self.success = status
        self.body = body


def success(body=''):
    """Make a successful GTP response."""
    return Response(status=True, body=body)


def error(body=''):
    """Make an error GTP response."""
    return Response(status=False, body=body)


def serialize(gtp_command, gtp_response):
    """Serialize a GTP response as a string.

    Needs the command we are responding to so we can match the sequence
    number.
    """
    return '%s%s %s\n\n' % (
        '=' if gtp_response.success else '?',
        '' if gtp_command.sequence is None else str(gtp_command.sequence),
        gtp_response.body,
    )
